/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 4/14/1993
 *-------------------------------------------------------------------
 */

#include <stdio.h>
#include "comm_int.h"

/*
 * Utilities and debugging routines for communication part
 */


void SYR_not_written(s)
char *s;
{
    fprintf(stderr, "*** Function %s has not been written yet ***", s);
    exit(1);
}

void SYR_panic(s)
char *s;
{
    fprintf(stderr, "PARTI internal error : %s\n", s);
    exit(1);
}

/*
 * Force nodes execute programs one after another. 
 */

void SYR_seq_begin()
{
    int i;

    for (i = 0; i <= SYR_myid; i++) {
        SYR_gsync();
    }
}

void SYR_seq_end()
{
    int i;

    for (i = SYR_myid; i < SYR_nproc; i++) {
        SYR_gsync();
    }
}

/*
 * Print out an array
 */

void SYR_dump_ary(ary, len)
int *ary;
int len;
{
    int i;

    for (i = 0; i < len; i++) {
        printf("%6d ", ary[i]);
    }
    printf("\n");
}

/*
 * print out a matrix
 */

void SYR_dump_mat(mat, row, cols)
int **mat;
int row, *cols;
{
    int i, j;

    for (i = 0; i < row; i++) {
        for (j = 0; j < cols[i]; j++) {
            printf("%6d ", mat[i][j]);
        }
        printf("\n    ");
    }
    printf("\n");
}

/*
 * Print out the contents of communication stage list
 */

void SYR_dump_comm_stage(mark)
SYR_comm_stage *mark;
{
    SYR_comm_stage *p;

    for (p = mark; p != NULL; p = p -> next) {
        printf("[%2d,%2d|%2d,%2d]",
            p -> s_partner, p -> s_len, p -> r_partner, p -> r_len);
    }
    printf("\n");
}

/*
 * Print out the contents of communication data structure
 */

void SYR_dump_com_ptn(ptn)
SYR_com_pattern *ptn;
{
    printf("n_msgs_to_recv=%d, n_msgs_to_send=%d, csched_type=%d\n",
        ptn -> n_msgs_to_recv, ptn -> n_msgs_to_send, ptn -> csched_type);
    printf("  tot_send_n=%d, tot_fetch_n=%d, g_n_stag=%d, s_n_stag=%d\n",
        ptn -> tot_send_n, ptn -> tot_fetch_n, ptn -> g_num_stages, ptn -> s_num_stages);
    printf("s_n:");
    SYR_dump_mat(&(ptn -> send_n), 1, &SYR_nproc);
    printf("s_l:");
    SYR_dump_mat(ptn -> send_list, SYR_nproc, ptn -> send_n);
    printf("f_n:");
    SYR_dump_mat(&(ptn -> fetch_n), 1, &SYR_nproc);
    printf("f_l:");
    SYR_dump_mat(ptn -> fetch_list, SYR_nproc, ptn -> fetch_n);
}

/*
 * Check amount of memory available.
 */

#define ONEK (1024)

int SYR_memavail()
{
    int alocunit, memsize;
    char *p1, *p2;

    memsize = 0;
    p1 = malloc(1);
    if (p1 != NULL) {
        alocunit = 1 << 30;
        while (alocunit >= ONEK) {
            memsize += alocunit;
            p2 = p1;
            p1 = realloc(p1, memsize);
            if (p1 == NULL) {
                memsize -= alocunit;
                p1 = p2;
            }
            alocunit >>= 1;
        }
        free(p1);
    }
    return(memsize);
}

/*
 * My LCG random number generator routines.
 * (This function can be used for internal SYR_ routines only)
 */

#define MODLUS 2147483647
#define MULT1       24112
#define MULT2       26147

static long SYR_rand_seed = 1;

/*
 * This function set the seed of internal random number generator
 * and return the current seed. The seed can be any integer value 
 * except for 0. If user want to set seed to be 0, the seed is
 * not changed.
 */

long SYR_srand(x)
long x;
{
    long old_seed;

    SYR_dbg_enter("SYR_srand");
    old_seed = SYR_rand_seed;
    SYR_rand_seed = (x == 0) ? old_seed : x;
    SYR_dbg_exit();
    return(old_seed);
}

double SYR_rand()
{
    long lowprd, hi31;

    SYR_dbg_enter("SYR_rand");
    lowprd = (SYR_rand_seed & 65535) * MULT1;
    hi31 = (SYR_rand_seed >> 16) * MULT1 + (lowprd >> 16);
    SYR_rand_seed = ((lowprd & 65535) - MODLUS) +
         ((hi31 & 32767) << 16) + (hi31 >> 15);
    if (SYR_rand_seed < 0) SYR_rand_seed += MODLUS;
    lowprd = (SYR_rand_seed & 65535) * MULT2;
    hi31 = (SYR_rand_seed >> 16) * MULT2 + (lowprd >> 16);
    SYR_rand_seed = ((lowprd & 65535) - MODLUS) +
         ((hi31 & 32767) << 16) + (hi31 >> 15);
    if (SYR_rand_seed < 0) SYR_rand_seed += MODLUS;
    SYR_dbg_exit();
    return ((double) ((SYR_rand_seed >> 7 | 1) + 1) / 16777216.0);
}
