/*-------------------------------------------------------------------
 * This file is written by Tseng-Hui Lin. If you have any question,
 * please send e-mail to:
 *     thlin@top.cis.syr.edu  or  thlin@nova.npac.syr.edu
 *
 * Last modify : 4/14/1993
 *-------------------------------------------------------------------
 */

/*
 * This file gives the name change for most communication system calls
 * This makes other files do not need to change when porting this 
 * system to other machines.
 */

#include <cube.h>
#include "port.h"

#define SYR_PORT_M_TYPE_BITS (20)        /* Leave 2 ^ 20 message type for user */
#define SYNC_TYPE_BASE (0X20000000)
static int myid, nproc, pid;
static int type_node;

/* General communication calls for all systems */
void SYR_port_init()
{
    myid = mynode();
    nproc = numnodes();
    pid = mypid();
    type_node = myid << SYR_PORT_M_TYPE_BITS;
}

int SYR_mynode()
{
    return(myid);
}

int SYR_numnodes()
{
    return(nproc);
}

void SYR_blk_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    type |= type_node;
    csend(type, buf, len, dest, pid);
    flushmsg(type, dest, pid);
}

void SYR_nonblk_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    csend(type | type_node, buf, len, dest, pid);
}

void SYR_sync_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    char sync_buf[1];

    /* Wait for receive node get ready */
    crecv(SYNC_TYPE_BASE + dest, sync_buf, 0);
    csend(type | type_node, buf, len, dest, pid);
}

int SYR_async_send(type, buf, len, dest)
int type, len, dest;
char *buf;
{
    return(isend(type | type_node, buf, len, dest, pid));
}

void SYR_recv(type, buf, len, src)
int type, len, src;
char *buf;
{
    type |= (src << SYR_PORT_M_TYPE_BITS);
    crecv(type, buf, len);
}

void SYR_sync_recv(type, buf, len, src)
int type, len, src;
char *buf;
{
    long msgid;
    char sync_buf[1];

    type |= (src << SYR_PORT_M_TYPE_BITS);
    msgid = irecv(type, buf, len);
    csend(SYNC_TYPE_BASE + myid, sync_buf, 0, src, pid);
    msgwait(msgid);
}

int SYR_async_recv(type, buf, len, src)
int type, len, src;
char *buf;
{
    type |= (src << SYR_PORT_M_TYPE_BITS);
    return(irecv(type, buf, len));
}

void SYR_send_recv(stype, sbuf, slen, dest, rtype, rbuf, rlen, src)
int stype, slen, dest, rtype, rlen, src;
char *sbuf, *rbuf;
{
    rtype |= (src << SYR_PORT_M_TYPE_BITS);
    csendrecv(stype | type_node, sbuf, slen, dest, pid,
        rtype, rbuf, rlen, src);
}

void SYR_sync_send_recv(stype, sbuf, slen, dest, rtype, rbuf, rlen, src)
int stype, slen, dest, rtype, rlen, src;
char *sbuf, *rbuf;
{
    long msgid;
    char sync_buf[1];

    rtype |= (src << SYR_PORT_M_TYPE_BITS);
    msgid = irecv(rtype, rbuf, rlen);
    csend(SYNC_TYPE_BASE + myid, sync_buf, 0, src, pid);

    /* Wait for receive node get ready */
    crecv(SYNC_TYPE_BASE + dest, sync_buf, 0);
    csend(stype | type_node, sbuf, slen, dest, pid);
    msgwait(msgid);
}

/* 
 * Special pairwise concurrent send and receive for iPSC 
 */

void SYR_swap(stype, sbuf, slen, dest, rtype, rbuf, rlen)
int stype, slen, dest, rtype, rlen;
char *sbuf, *rbuf;
{
    long msgid;
    int i;
    char sync_buf[1];

    rtype |= (dest << SYR_PORT_M_TYPE_BITS);
    stype |= type_node;
    msgid = irecv(rtype, rbuf, rlen);
    if (myid > dest) {
        /* A pair of send/recv to make sure the links */
        /* are useable at this time                   */
        crecv(SYNC_TYPE_BASE + dest, sync_buf, 0);
        csend(SYNC_TYPE_BASE + myid, sync_buf, 0, dest, pid);
        /* synchronous with the partner */
        crecv(SYNC_TYPE_BASE + dest, sync_buf, 0);
    }
    else {
        /* A pair of send/recv to make sure the links */
        /* are useable at this time                   */
        csend(SYNC_TYPE_BASE + myid, sync_buf, 0, dest, pid);
        crecv(SYNC_TYPE_BASE + dest, sync_buf, 0);
        /* synchronous with the partner */
        csend(SYNC_TYPE_BASE + myid, sync_buf, 0, dest, pid);
        /* delay for 12.5 us. It makes the following */
        /* send of both partner start sending almost */
        /* at the same time. It's very important for */
        /* concurrent sending/receiving on iPSC.     */
        for (i = 0; i < 100; i++);
    }
    csend(stype, sbuf, slen, dest, pid);
    msgwait(msgid);
}

int SYR_recv_len()
{
    return(infocount());
}

int SYR_recv_sender()
{
    return(infonode());
}

int SYR_recv_type()
{
    return(infotype());
}

void SYR_async_msg_wait(msgid)
int msgid;
{
    msgwait(msgid);
}

int SYR_async_msg_done(msgid)
int msgid;
{
    return(msgdone(msgid));
}

void SYR_concat(elem, elem_len, whole)
char *elem, *whole;
int elem_len;
{
    int work;
    gcol(elem, elem_len, whole, elem_len * nproc, &work);
}

/* Global synchronous operation */

void SYR_gsync()
{
    gsync();
}

/* Time routines */

double SYR_clock()
{
    return(dclock());
}

/*----------------------------------------------------------------*/
/* Specific communication calls for iPSC */

int SYR_mypid()
{
    return(pid);
}

int SYR_nodedim()
{
    return(nodedim());
}

